# -*- coding: utf-8 -*-

from yapconf.docs import build_markdown_table


# flake8: noqa
def test_build_markdown_table():
    headers = {'foo': 'Foo', 'bar': 'Bar'}
    rows = [
        {
            'bar': 'bar_value',
            'foo': 'x',
        },
        {
            'bar': 'bar_value2',
            'foo': 'y',
        },
    ]
    table = build_markdown_table(headers, rows, ['foo', 'bar'])
    assert table == ("| Foo | Bar        |\n"
                     "| --- | ---------- |\n"
                     "| x   | bar_value  |\n"
                     "| y   | bar_value2 |\n")


def test_generate_markdown_doc_simple(simple_spec):
    doc = simple_spec.generate_documentation('My App Name')
    assert doc == (
"""# My App Name Configuration

This document describes the configuration for My App Name. Each section will 
document a particular configuration value and its description. First, 
though, we start with the possible sources. 

This documentation was auto-generated by [yapconf.](https://github.com/loganasherjones/yapconf)


## Sources

My App Name configuration can be loaded from the below sources:

## Configuration

This section outlines the various configuration items My App Name supports.

| Name                      | Type    | Default | Description |
| ------------------------- | ------- | ------- | ----------- |
| [my_bool](#my_bool)       | bool    | None    | None        |
| [my_complex](#my_complex) | complex | None    | None        |
| [my_float](#my_float)     | float   | None    | None        |
| [my_int](#my_int)         | int     | None    | None        |
| [my_long](#my_long)       | long    | None    | None        |
| [my_string](#my_string)   | str     | None    | None        |

### my_bool

No description provided.

| Attribute     | Value       |
| ------------- | ----------- |
| **item_type** | `bool`      |
| **default**   | `None`      |
| **env_name**  | `MY_BOOL`   |
| **required**  | `True`      |
| **cli_name**  | `--my-bool` |
| **fallback**  | `None`      |
| **choices**   | `None`      |

You can set my_bool from the environment by setting the environment variable `MY_BOOL`

You can set `my_bool` from the command-line by specifying `--my-bool` at My App Name's entrypoint.



### my_complex

No description provided.

| Attribute     | Value          |
| ------------- | -------------- |
| **item_type** | `complex`      |
| **default**   | `None`         |
| **env_name**  | `MY_COMPLEX`   |
| **required**  | `True`         |
| **cli_name**  | `--my-complex` |
| **fallback**  | `None`         |
| **choices**   | `None`         |

You can set my_complex from the environment by setting the environment variable `MY_COMPLEX`

You can set `my_complex` from the command-line by specifying `--my-complex` at My App Name's entrypoint.



### my_float

No description provided.

| Attribute     | Value        |
| ------------- | ------------ |
| **item_type** | `float`      |
| **default**   | `None`       |
| **env_name**  | `MY_FLOAT`   |
| **required**  | `True`       |
| **cli_name**  | `--my-float` |
| **fallback**  | `None`       |
| **choices**   | `None`       |

You can set my_float from the environment by setting the environment variable `MY_FLOAT`

You can set `my_float` from the command-line by specifying `--my-float` at My App Name's entrypoint.



### my_int

No description provided.

| Attribute     | Value      |
| ------------- | ---------- |
| **item_type** | `int`      |
| **default**   | `None`     |
| **env_name**  | `MY_INT`   |
| **required**  | `True`     |
| **cli_name**  | `--my-int` |
| **fallback**  | `None`     |
| **choices**   | `None`     |

You can set my_int from the environment by setting the environment variable `MY_INT`

You can set `my_int` from the command-line by specifying `--my-int` at My App Name's entrypoint.



### my_long

No description provided.

| Attribute     | Value       |
| ------------- | ----------- |
| **item_type** | `long`      |
| **default**   | `None`      |
| **env_name**  | `MY_LONG`   |
| **required**  | `True`      |
| **cli_name**  | `--my-long` |
| **fallback**  | `None`      |
| **choices**   | `None`      |

You can set my_long from the environment by setting the environment variable `MY_LONG`

You can set `my_long` from the command-line by specifying `--my-long` at My App Name's entrypoint.



### my_string

No description provided.

| Attribute     | Value         |
| ------------- | ------------- |
| **item_type** | `str`         |
| **default**   | `None`        |
| **env_name**  | `MY_STRING`   |
| **required**  | `True`        |
| **cli_name**  | `--my-string` |
| **fallback**  | `None`        |
| **choices**   | `None`        |

You can set my_string from the environment by setting the environment variable `MY_STRING`

You can set `my_string` from the command-line by specifying `--my-string` at My App Name's entrypoint.


"""
    )



# -*- coding: utf-8 -*-
import os

import pytest

import yapconf
from yapconf import YapconfSpec
from yapconf.items import from_specification


@pytest.fixture
def simple_item_spec():
    """A specification for a string item."""
    return {'foo': {}}


@pytest.fixture
def bool_item_spec():
    """A specification for a boolean item."""
    return {'my_bool': {'type': 'bool', 'required': True}}


@pytest.fixture
def unformatted_item_spec():
    """A specification for an unformatted string item."""
    return {
        'weirdName-Cool_stuff lol': {
            'required': True,
            'format_env': False,
            'format_cli': False,
        }
    }


@pytest.fixture
def unformatted_bool_item_spec():
    """A specification for an unformatted boolean item."""
    return {
        'weirdName-Cool_stuff lol': {
            'type': 'bool',
            'required': True,
            'format_env': False,
            'format_cli': False,
        }
    }


@pytest.fixture
def list_item_spec(simple_item_spec):
    """A specification for a list item."""
    return {
        'foos': {
            'required': True,
            'items': {'simple_item': simple_item_spec}
        }
    }


@pytest.fixture
def bool_list_item_spec(bool_item_spec):
    """A specification for a list of boolean items."""
    return {
        'my_bools': {
            'required': True,
            'items': bool_item_spec
        }
    }


@pytest.fixture
def simple_item(simple_item_spec):
    """A YacponfItem for a string config value."""
    return from_specification(simple_item_spec)['foo']


@pytest.fixture
def bool_item(bool_item_spec):
    """A YacponfItem for a boolean config value."""
    return from_specification(bool_item_spec)['my_bool']


@pytest.fixture
def unformatted_item(unformatted_item_spec):
    """A YapconfItem for an unformatted string value."""
    items = from_specification(unformatted_item_spec)
    return items['weirdName-Cool_stuff lol']


@pytest.fixture
def unformatted_bool_item(unformatted_bool_item_spec):
    """A YapconfItem for an unformatted boolean value."""
    # unformatted_item_spec['type'] = 'bool'
    items = from_specification(unformatted_bool_item_spec)
    return items['weirdName-Cool_stuff lol']


@pytest.fixture
def list_item(simple_item_spec):
    """A YapconfItem for a list value."""
    return from_specification({
        'foos': {
            'type': 'list',
            'items': {
                'simple_item': simple_item_spec
            }
        }
    })['foos']


@pytest.fixture
def bool_list_item(bool_item_spec):
    """A YapconfItem for a list of boolean values."""
    return from_specification({
        'my_bools': {
            'type': 'list',
            'items': bool_item_spec
        }
    })['my_bools']


@pytest.fixture
def dict_item(simple_item_spec):
    """A YapconfItem for a dictionary."""
    return from_specification({
        'foo_dict': {
            'type': 'dict',
            'items': {'foo': simple_item_spec}
        }
    })['foo_dict']


@pytest.fixture
def db_item():
    """A YapconfItem for a dictionary that represents a db config."""
    return from_specification(
        {
            'db': {
                'type': 'dict',
                'items': {
                    'name': {
                        'required': True,
                        'alt_env_names': ['ALT_NAME'],
                    },
                    'port': {
                        'required': True,
                        'type': 'int',
                    },
                    'verbose': {
                        'required': True,
                        'type': 'bool',
                    },
                    'users': {
                        'type': 'list',
                        'required': True,
                        'items': {
                            'user': {
                                'required': True,
                                'type': 'str',
                            },
                        },
                    },
                    'log': {
                        'type': 'dict',
                        'required': True,
                        'items': {
                            'level': {
                                'required': True,
                                'type': 'str'
                            },
                            'file': {
                                'required': True,
                                'type': 'str'
                            },
                        },
                    },
                },
            },
        },
    )['db']


@pytest.fixture
def basic_spec(simple_item_spec):
    """Most basic spec you can have"""
    return YapconfSpec(simple_item_spec)


@pytest.fixture
def simple_spec():
    """Simple YapconfSpec for all YapconfItem variations"""
    return YapconfSpec(
        {
            'my_string': {'type': 'str', 'required': True, },
            'my_int': {'type': 'int', 'required': True, },
            'my_long': {'type': 'long', 'required': True, },
            'my_float': {'type': 'float', 'required': True, },
            'my_bool': {'type': 'bool', 'required': True, },
            'my_complex': {'type': 'complex', 'required': True, },
        }
    )


@pytest.fixture
def spec_with_lists():
    """YapconfSpec for testing YapconfListItem variations"""
    return YapconfSpec(
        {
            'simple_list': {
                'type': 'list',
                'required': True,
                'items': {
                    'list_item': {
                        'type': 'str', 'required': True
                    }
                }
            },
            'top_list': {
                'type': 'list',
                'required': True,
                'items': {
                    'nested_list': {
                        'type': 'list',
                        'required': True,
                        'items': {
                            'nested_list_items': {
                                'type': 'int',
                                'required': True
                            }
                        }
                    }
                }
            },
            'list_of_dictionaries': {
                'type': 'list',
                'required': True,
                'items': {
                    'list_item': {
                        'type': 'dict',
                        'required': True,
                        'items': {
                            'foo': {'type': 'str', 'required': True, },
                            'bar': {'type': 'str', 'required': False, }
                        }
                    }
                }
            }
        }
    )


@pytest.fixture
def spec_with_dicts():
    """YapconfSpec for testing YapconfDictItem variations"""
    return YapconfSpec({
        'database': {
            'type': 'dict',
            'required': True,
            'items': {
                'name': {'type': 'str', 'required': True, },
                'host': {'type': 'str', 'required': True, },
                'port': {'type': 'int', 'required': True, },
            }
        },
        'foo': {
            'type': 'dict',
            'required': True,
            'items': {
                'bar': {
                    'type': 'dict',
                    'required': True,
                    'items': {
                        'baz': {
                            'type': 'str', 'required': True,
                        },
                    },
                },
                'bat': {
                    'type': 'bool',
                }
            },
        },
    })


@pytest.fixture
def real_world_spec():
    """YapconfSpec based on a 'real-world' example"""
    current_dir = os.path.abspath(os.path.dirname(__file__))
    filename = os.path.join(current_dir, 'files', 'real_world', 'spec.yaml')
    return YapconfSpec(filename, file_type='yaml', env_prefix='MY_APP_')


@pytest.fixture
def current_config():
    current_dir = os.path.abspath(os.path.dirname(__file__))
    filename = os.path.join(
        current_dir,
        'files',
        'real_world',
        'current_config.yaml'
    )
    return yapconf.load_file(filename, 'yaml')


@pytest.fixture
def previous_config():
    current_dir = os.path.abspath(os.path.dirname(__file__))
    filename = os.path.join(
        current_dir,
        'files',
        'real_world',
        'previous_config.yaml'
    )
    return yapconf.load_file(filename, 'yaml')


@pytest.fixture
def default_current_config():
    current_dir = os.path.abspath(os.path.dirname(__file__))
    filename = os.path.join(
        current_dir,
        'files',
        'real_world',
        'default_current_config.yaml'
    )
    return yapconf.load_file(filename, 'yaml')


@pytest.fixture
def default_previous_config():
    current_dir = os.path.abspath(os.path.dirname(__file__))
    filename = os.path.join(
        current_dir,
        'files',
        'real_world',
        'default_previous_config.yaml'
    )
    return yapconf.load_file(filename, 'yaml')


@pytest.fixture
def example_spec():
    return YapconfSpec(
        {
            'foo': {},
            'emoji': {},
            u'💩': {},
            'db': {
                'type': 'dict',
                'items': {
                    'name': {},
                    'port': {'type': 'int'}
                }
            },
            'items': {
                'type': 'list',
                'items': {
                    'item': {'type': 'int'},
                },
            },
        }
    )


@pytest.fixture
def example_data():
    return {
        "foo": "bar",
        "emoji": u"💩",
        u"💩": u"🐍",
        "db": {
            "name": "db_name",
            "port": 123
        },
        "items": [1, 2, 3]
    }


@pytest.fixture
def fallback_spec():
    return YapconfSpec({
        'defaults': {
            'type': 'dict',
            'items': {
                'str': {'type': 'str', 'default': 'default_str'},
                'int': {'type': 'int', 'default': 123},
                'long': {'type': 'long', 'default': 123},
                'float': {'type': 'float', 'default': 123.123},
                'bool': {'type': 'bool', 'default': True},
                'complex': {'type': 'complex', 'default': 1j},
                'list': {
                    'type': 'list',
                    'default': [1, 2, 3],
                    'items': {
                        'list_item': {'type': 'int', 'default': 1}
                    },
                },
                'dict': {
                    'type': 'dict',
                    'default': {'foo': 'dict_default'},
                    'items': {
                        'foo': {'type': 'str', 'default': 'item_default'}
                    },
                },
            },
        },
        'str': {'type': 'str', 'fallback': 'defaults.str'},
        'int': {'type': 'int', 'fallback': 'defaults.int'},
        'long': {'type': 'long', 'fallback': 'defaults.long'},
        'float': {'type': 'float', 'fallback': 'defaults.float'},
        'bool': {'type': 'bool', 'fallback': 'defaults.bool'},
        'complex': {'type': 'complex', 'fallback': 'defaults.complex'},
        'list': {
            'type': 'list',
            'fallback': 'defaults.list',
            'items': {
                'list_item': {
                    'type': 'int',
                    'fallback': 'defaults.list.list_item'
                },
            },
        },
        'dict': {
            'type': 'dict',
            'fallback': 'defaults.dict',
            'items': {
                'foo': {
                    'type': 'str',
                    'fallback': 'defaults.dict.foo'
                },
            },
        },
    })



# -*- coding: utf-8 -*-
import os

import pytest
from box import Box
from mock import patch

import yapconf


@pytest.fixture
def ascii_data():
    return {
        'foo': 'bar',
        'db': {'name': 'db_name', 'port': 123},
        'items': [1, 2, 3]
    }


@pytest.fixture
def unicode_data():
    return {
        u'foo': u'bar',
        'db': {'name': 'db_name', 'port': 123},
        'items': [1, 2, 3],
        u'💩': u'🐍',
        'emoji': u'💩'
    }


@pytest.mark.parametrize('orig,expected', [
    ('CamelCase', 'camel_case'),
    ('CamelCamelCase', 'camel_camel_case'),
    ('Camel2Camel2Case', 'camel2_camel2_case'),
    ('getHTTPResponseCode', 'get_http_response_code'),
    ('get2HTTPResponseCode', 'get2_http_response_code'),
    ('HTTPResponseCode', 'http_response_code'),
    ('HTTPResponseCodeXYZ', 'http_response_code_xyz'),
    ('snake-case', 'snake_case'),
    ('snake-snake-case', 'snake_snake_case'),
    ('snake2-snake2-case', 'snake2_snake2_case'),
    ('snake3-snake3_case', 'snake3_snake3_case'),
    (
        ' CamelGetHTTPResponse_code_snake2_case is a pain',
        'camel_get_http_response_code_snake2_case_is_a_pain'
    ),
])
def test_convert_to_snake(orig, expected):
    assert expected == yapconf.change_case(orig, separator='_')


@pytest.mark.parametrize('orig,expected', [
    ('CamelCase', 'camel-case'),
    ('CamelCamelCase', 'camel-camel-case'),
    ('Camel2Camel2Case', 'camel2-camel2-case'),
    ('getHTTPResponseCode', 'get-http-response-code'),
    ('get2HTTPResponseCode', 'get2-http-response-code'),
    ('HTTPResponseCode', 'http-response-code'),
    ('HTTPResponseCodeXYZ', 'http-response-code-xyz'),
    ('snake_case', 'snake-case'),
    ('snake_snake_case', 'snake-snake-case'),
    ('snake2_snake2_case', 'snake2-snake2-case'),
    ('snake3_snake3-case', 'snake3-snake3-case'),
    (' CamelGetHTTPResponse_code_snake2_case is a pain',
     'camel-get-http-response-code-snake2-case-is-a-pain')
])
def test_convert_to_kebab(orig, expected):
    assert expected == yapconf.change_case(orig)


@pytest.mark.parametrize('filename,file_type,expected', [
    (
        'ascii.yaml',
        'yaml',
        pytest.lazy_fixture('ascii_data')
    ),
    (
        'unicode.yaml',
        'yaml',
        pytest.lazy_fixture('unicode_data')
    ),
    (
        'ascii.json',
        'json',
        pytest.lazy_fixture('ascii_data')
    ),
    (
        'unicode.json',
        'json',
        pytest.lazy_fixture('unicode_data')
    ),
])
def test_load_file(filename, file_type, expected):
    current_dir = os.path.abspath(os.path.dirname(__file__))
    full_path = os.path.join(current_dir, 'files', filename)
    data = yapconf.load_file(full_path, file_type=file_type)
    assert data == expected


@pytest.mark.parametrize('filename,file_type', [
    ('bad.json', 'json'),
    ('ascii.json', 'INVALID_TYPE'),
])
def test_load_file_error(filename, file_type):
    current_dir = os.path.abspath(os.path.dirname(__file__))
    full_path = os.path.join(current_dir, 'files', filename)
    with pytest.raises(ValueError):
        yapconf.load_file(full_path,
                          file_type=file_type,
                          klazz=ValueError)


@pytest.mark.parametrize('data,file_type', [
    (
        pytest.lazy_fixture('ascii_data'),
        'yaml'
    ),
    (
        pytest.lazy_fixture('unicode_data'),
        'yaml'
    ),
    (
        pytest.lazy_fixture('unicode_data'),
        'json'
    ),
    (
        pytest.lazy_fixture('ascii_data'),
        'json'
    )
])
def test_dump_data(tmpdir, data, file_type):
    path = tmpdir.join('test.%s' % file_type)
    filename = os.path.join(path.dirname, path.basename)
    yapconf.dump_data(data, filename, file_type)
    assert data == yapconf.load_file(filename, file_type)


def test_dump_box(ascii_data):
    data = {'writes': ''}

    def mock_write(x):
        data['writes'] += x

    expected = (
        'db:{n}'
        '  name: db_name{n}'
        '  port: 123{n}'
        'foo: bar{n}'
        'items:{n}'
        '- 1{n}'
        '- 2{n}'
        '- 3{n}'
    ).format(n=os.linesep)

    boxed_data = Box(ascii_data)
    with patch('sys.stdout.write', mock_write):
        yapconf.dump_data(boxed_data, file_type='yaml')

    assert data['writes'] == expected


@pytest.mark.parametrize('original,expected', [
    ({}, {}),
    ({'foo': 'bar'}, {'foo': 'bar'}),
    ({'foo': {'bar': 'baz'}}, {'foo.bar': 'baz'}),
    (
        {'foo': {'bar': {'baz': 'bat', 'bax': 'bat'}, 'bat': 'bar'}},
        {'foo.bar.baz': 'bat', 'foo.bar.bax': 'bat', 'foo.bat': 'bar'}
    ),
    (
        {
            'list': [
                {'foo': 'foo_value', 'bar': 'bar_value'},
                {'foo': 'foo_value2', 'bar': 'bar_value2'},
            ]
        },
        {
            'list': [
                {'list.foo': 'foo_value', 'list.bar': 'bar_value'},
                {'list.foo': 'foo_value2', 'list.bar': 'bar_value2'}
            ]
        }
    ),
    (
        {
            'foo': {'list1': [1], 'list2': [{'bar': 'baz'}]}
        },
        {
            'foo.list1': [1], 'foo.list2': [{'foo.list2.bar': 'baz'}]
        }
    ),
    (
        {
            'foo': {
                'list1': [
                    {
                        'bar': {
                            'list2': [1, 2, 3],
                            'bar2': 'bar2_value',
                            'bat': {
                                'bat2': 'bat2_value',
                                'bat_list_dict': [
                                    {
                                        'baz': 'baz_value',
                                        'bazl': [4, 5, 6],
                                        'bazd': {'bazd_key': 'bazd_value'}
                                    }
                                ]
                            }
                        }
                    }
                ]
            }
        },
        {
            'foo.list1': [
                {
                    'foo.list1.bar.list2': [1, 2, 3],
                    'foo.list1.bar.bar2': 'bar2_value',
                    'foo.list1.bar.bat.bat2': 'bat2_value',
                    'foo.list1.bar.bat.bat_list_dict': [
                        {
                            'foo.list1.bar.bat.bat_list_dict.baz': 'baz_value',
                            'foo.list1.bar.bat.bat_list_dict.bazl': [4, 5, 6],
                            'foo.list1.bar.bat.bat_list_dict.bazd.bazd_key':
                                'bazd_value'
                        }
                    ]

                }
            ]
        }
    )
])
def test_flatten(original, expected):
    assert yapconf.flatten(original) == expected



# -*- coding: utf-8 -*-
import logging
import os
import sys

import six

import yapconf
from box import Box
from yapconf.docs import generate_markdown_doc
from yapconf.exceptions import (YapconfItemNotFound, YapconfLoadError,
                                YapconfSourceError, YapconfSpecError)
from yapconf.handlers import ConfigChangeHandler
from yapconf.items import YapconfDictItem, YapconfListItem, from_specification
from yapconf.sources import get_source

if sys.version_info.major < 3:
    from io import open


class YapconfSpec(object):
    """Object which holds your configuration's specification.

    The YapconfSpec item is the main interface into the yapconf package.
    It will help you load, migrate, update and add arguments for your
    application.

    Examples:
        >>> from yapconf import YapconfSpec

        First define a specification

        >>> my_spec = YapconfSpec(
        ...   {"foo": {"type": "str", "default": "bar"}},
        ...   env_prefix='MY_APP_'
        ... )

        Then load the configuration in whatever order you want!
        load_config will automatically look for the 'foo' value in
        '/path/to/config.yml', then the environment, finally
        falling back to the default if it was not found elsewhere

        >>> config = my_spec.load_config('/path/to/config.yml', 'ENVIRONMENT')
        >>> print(config.foo)
        >>> print(config['foo'])

    """

    def __init__(self, specification, file_type='json', env_prefix=None,
                 encoding='utf-8', separator='.'):
        self._file_type = file_type
        self._encoding = encoding

        if self._file_type not in yapconf.FILE_TYPES:
            raise YapconfSpecError(
                'Unsupported file type: {0}. Supported file types are: {1}'
                .format(self._file_type, yapconf.FILE_TYPES)
            )

        self._specification = self._load_specification(specification)
        self._env_prefix = env_prefix
        self._separator = separator
        self._yapconf_items = from_specification(self._specification,
                                                 self._env_prefix,
                                                 self._separator)
        self._logger = logging.getLogger(__name__)
        self._sources = {}

    def _load_specification(self, specification):
        if isinstance(specification, six.string_types):
            specification = yapconf.load_file(
                specification,
                file_type=self._file_type,
                klazz=YapconfSpecError
            )

        if not isinstance(specification, dict):
            raise YapconfSpecError(
                'Specification must be a dictionary or a filename which '
                'contains a loadable dictionary. Supported file types are {0}'
                .format(yapconf.FILE_TYPES)
            )

        self._validate_specification(specification)
        return specification

    def _validate_specification(self, specification):
        for item_name, item_info in six.iteritems(specification):
            if not isinstance(item_info, dict):
                raise YapconfSpecError("Invalid specification. {0} is not a "
                                       "dictionary.".format(item_name))

            item_type = item_info.get('type', 'str')
            nested_items = item_info.get('items', {})
            if nested_items and item_type not in ['list', 'dict']:
                raise YapconfSpecError("Invalid specification. {0} is a {1} "
                                       "type item, but it has children. "
                                       "Maybe you meant to set type to either"
                                       "'list' or 'dict'?"
                                       .format(item_name, item_type))
            elif item_type in ['list', 'dict'] and not nested_items:
                raise YapconfSpecError("Invalid specification. {0} is a {1} "
                                       "type item, but it has no children. "
                                       "This is not allowed."
                                       .format(item_name, item_type))
            elif item_type in ['list', 'dict']:
                self._validate_specification(nested_items)

    @property
    def items(self):
        return self._yapconf_items

    @property
    def sources(self):
        return self._sources

    @property
    def defaults(self):
        """dict: All defaults for items in the specification."""
        return self._get_defaults(self._yapconf_items.values())

    def _get_defaults(self, items):
        defaults = {}
        for item in items:
            if isinstance(item, YapconfDictItem):
                defaults[item.name] = self._get_defaults(
                    item.children.values()
                )
            else:
                defaults[item.name] = item.default
        return defaults

    def add_arguments(self, parser, bootstrap=False):
        """Adds all items to the parser passed in.

        Args:
            parser (argparse.ArgumentParser): The parser to add all items to.
            bootstrap (bool): Flag to indicate whether you only want to mark
                bootstrapped items as required on the command-line.

        """
        [item.add_argument(parser, bootstrap)
         for item in self._get_items(bootstrap=False)]

    def add_source(self, label, source_type, **kwargs):
        """Add a source to the spec.

        Sources should have a unique label. This will help tracing where your
        configurations are coming from if you turn up the log-level.

        The keyword arguments are significant. Different sources require
        different keyword arguments. Required keys for each source_type are
        listed below, for a detailed list of all possible arguments, see the
        individual source's documentation.

        source_type: dict
            required keyword arguments:
                - data - A dictionary

        source_type: environment
            No required keyword arguments.

        source_type: etcd
            required keyword arguments:
                - client - A client from the python-etcd package.

        source_type: json
            required keyword arguments:
                - filename - A JSON file.
                - data - A string representation of JSON

        source_type: kubernetes
            required keyword arguments:
                - client - A client from the kubernetes package
                - name - The name of the ConfigMap to load

        source_type: yaml
            required keyword arguments:
                - filename - A YAML file.

        Args:
            label (str): A label for the source.
            source_type (str): A source type, available source types depend
            on the packages installed. See ``yapconf.ALL_SUPPORTED_SOURCES``
            for a complete list.

        """
        self._sources[label] = get_source(label, source_type, **kwargs)

    def find_item(self, fq_name):
        """Find an item in the specification by fully qualified name.

        Args:
            fq_name (str): Fully-qualified name of the item.

        Returns:
            The item if it is in the specification. None otherwise

        """
        names = fq_name.split(self._separator)
        current = self._yapconf_items
        for name in names:
            if isinstance(current, (YapconfDictItem, YapconfListItem)):
                current = current.children

            if name not in current:
                return None
            current = current[name]
        return current

    def get_item(self, name, bootstrap=False):
        """Get a particular item in the specification.

        Args:
            name (str): The name of the item to retrieve.
            bootstrap (bool): Only search bootstrap items

        Returns (YapconfItem):
            A YapconfItem if it is found, None otherwise.

        """
        for item in self._get_items(bootstrap):
            if item.name == name:
                return item
        return None

    def update_defaults(self, new_defaults, respect_none=False):
        """Update items defaults to the values in the new_defaults dict.

        Args:
            new_defaults (dict): A key-value pair of new defaults to be
                applied.
            respect_none (bool): Flag to indicate if ``None`` values should
                constitute an update to the default.

        """
        for key, value in six.iteritems(new_defaults):
            item = self.get_item(key)
            if item is None:
                raise YapconfItemNotFound("Cannot update default for {0}, "
                                          "there is no config item by the "
                                          "name of {1}".format(key, key), None)

            item.update_default(value, respect_none)

    def generate_documentation(self, app_name, **kwargs):
        """Generate documentation for this specification.

        Documentation is generated in Markdown format. An example
        of the generated documentation can be found at:

        https://github.com/loganasherjones/yapconf/blob/master/example/doc.md

        Args:
            app_name (str): The name of your application.

        Keyword Args:
            output_file_name (str): If provided, will write to this file.
            encoding (str): The encoding to use for the output file. Default
            is utf-8.

        Returns:
            A string representation of the documentation.

        """
        output_file = kwargs.get('output_file_name')
        encoding = kwargs.get('encoding', 'utf-8')
        doc_string = generate_markdown_doc(app_name, self)
        if output_file:
            with open(output_file, 'w', encoding=encoding) as doc_file:
                doc_file.write(doc_string)

        return doc_string

    def load_filtered_config(self, *args, **kwargs):
        """Loads a filtered version of the configuration.

        The order of arguments is the same as `load_config`.

        Keyword Args:
            bootstrap: If set to true, indicates only bootstrap items should
                be loaded.
            exclude_bootstrap: If set to true, load all non-bootstrapped items.
            include: A list of fully qualified key names that should be
                included. Only these items will be included in the config that
                is returned.
            exclude: A list of fully qualified key names that should be
                excluded. If there are conflicts in which items should be
                included, all items in include are guaranteed to be included,
                otherwise they will be excluded.

        """
        overrides = self._generate_overrides(*args)
        filtered_items = {}
        for item_name, item in six.iteritems(self._yapconf_items):
            result = item.apply_filter(**kwargs)
            if result is not None:
                filtered_items[item_name] = result

        return Box({
            item.name: item.get_config_value(overrides) for item in
            filtered_items.values()
        })

    def load_config(self, *args, **kwargs):
        """Load a config based on the arguments passed in.

        The order of arguments passed in as *args is significant. It indicates
        the order of precedence used to load configuration values. Each
        argument can be a string, dictionary or a tuple. There is a special
        case string called 'ENVIRONMENT', otherwise it will attempt to load the
        filename passed in as a string.

        By default, if a string is provided, it will attempt to load the
        file based on the file_type passed in on initialization. If you
        want to load a mixture of json and yaml files, you can specify them
        as the 3rd part of a tuple.

        Examples:
            You can load configurations in any of the following ways:

            >>> my_spec = YapconfSpec({'foo': {'type': 'str'}})
            >>> my_spec.load_config('/path/to/file')
            >>> my_spec.load_config({'foo': 'bar'})
            >>> my_spec.load_config('ENVIRONMENT')
            >>> my_spec.load_config(('label', {'foo': 'bar'}))
            >>> my_spec.load_config(('label', '/path/to/file.yaml', 'yaml'))
            >>> my_spec.load_config(('label', '/path/to/file.json', 'json'))

            You can of course combine each of these and the order will be
            held correctly.

        Args:
            *args:
            **kwargs: The only supported keyword argument is 'bootstrap'
                which will indicate that only bootstrap configurations
                should be loaded.

        Returns:
            box.Box: A Box object which is subclassed from dict. It should
                behave exactly as a dictionary. This object is guaranteed to
                contain at least all of your required configuration items.

        Raises:
            YapconfLoadError: If we attempt to load your args and something
                goes wrong.
            YapconfItemNotFound: If an item is required but could not be found
                in the configuration.
            YapconfItemError: If a possible value was found but the type
                cannot be determined.
            YapconfValueError: If a possible value is found but during
                conversion, an exception was raised.
        """
        bootstrap = kwargs.get('bootstrap', False)
        overrides = self._generate_overrides(*args)
        config = self._generate_config_from_overrides(overrides, bootstrap)
        return Box(config)

    def spawn_watcher(self, label, target=None, eternal=False):
        """Spawns a config watcher in a separate daemon thread.

        If a particular config value changes, and the item has a
        ``watch_target`` defined, then that method will be called.

        If a ``target`` is passed in, then it will call the ``target``
        anytime the config changes.

        Args:
            label (str): Should match a label added through ``add_source``
            target (func): Should be a function that takes two arguments,
            the old configuration and the new configuration.
            eternal (bool): Determines if watcher threads should be restarted
            if they die.

        Returns:
            The thread that was spawned.

        """

        if label not in self._sources:
            raise YapconfSourceError(
                'Cannot watch %s no source named %s' % (label, label)
            )

        current_config = self._sources[label].get_data()
        handler = ConfigChangeHandler(current_config, self, target)
        return self._sources[label].watch(handler, eternal)

    def migrate_config_file(
        self,
        config_file_path,
        always_update=False,
        current_file_type=None,
        output_file_name=None,
        output_file_type=None,
        create=True,
        update_defaults=True,
        dump_kwargs=None,
        include_bootstrap=True,
    ):
        """Migrates a configuration file.

        This is used to help you update your configurations throughout the
        lifetime of your application. It is probably best explained through
        example.

        Examples:
            Assume we have a JSON config file ('/path/to/config.json')
            like the following:
            ``{"db_name": "test_db_name", "db_host": "1.2.3.4"}``

            >>> spec = YapconfSpec({
            ...    'db_name': {
            ...        'type': 'str',
            ...        'default': 'new_default',
            ...        'previous_defaults': ['test_db_name']
            ...    },
            ...    'db_host': {
            ...        'type': 'str',
            ...        'previous_defaults': ['localhost']
            ...    }
            ... })

            We can migrate that file quite easily with the spec object:

            >>> spec.migrate_config_file('/path/to/config.json')

            Will result in /path/to/config.json being overwritten:
            ``{"db_name": "new_default", "db_host": "1.2.3.4"}``

        Args:
            config_file_path (str): The path to your current config
            always_update (bool): Always update values (even to None)
            current_file_type (str): Defaults to self._file_type
            output_file_name (str): Defaults to the current_file_path
            output_file_type (str): Defaults to self._file_type
            create (bool): Create the file if it doesn't exist (otherwise
                error if the file does not exist).
            update_defaults (bool): Update values that have a value set to
                something listed in the previous_defaults
            dump_kwargs (dict): A key-value pair that will be passed to dump
            include_bootstrap (bool): Include bootstrap items in the output

        Returns:
            box.Box: The newly migrated configuration.
        """

        current_file_type = current_file_type or self._file_type
        output_file_type = output_file_type or self._file_type
        output_file_name = output_file_name or config_file_path

        current_config = self._get_config_if_exists(config_file_path,
                                                    create,
                                                    current_file_type)

        migrated_config = {}

        if include_bootstrap:
            items = self._yapconf_items.values()
        else:
            items = [
                item for item in self._yapconf_items.values()
                if not item.bootstrap
            ]
        for item in items:
            item.migrate_config(current_config, migrated_config,
                                always_update, update_defaults)

        if create:
            yapconf.dump_data(migrated_config,
                              filename=output_file_name,
                              file_type=output_file_type,
                              klazz=YapconfLoadError,
                              dump_kwargs=dump_kwargs)

        return Box(migrated_config)

    def _get_config_if_exists(self, config_file_path, create, file_type):
        if not os.path.isfile(config_file_path) and create:
            return {}
        elif not os.path.isfile(config_file_path):
            raise YapconfLoadError("Error migrating config file {0}. "
                                   "File does not exist. If you would like "
                                   "to create the file, you need to pass "
                                   "the create flag.".format(config_file_path))

        return yapconf.load_file(config_file_path,
                                 file_type=file_type,
                                 klazz=YapconfLoadError)

    def _get_items(self, bootstrap):
        if not bootstrap:
            return self._yapconf_items.values()
        else:
            return [item for item in self._yapconf_items.values()
                    if item.bootstrap]

    def _generate_config_from_overrides(self, overrides, bootstrap):
        return {
            item.name: item.get_config_value(overrides) for item in
            self._get_items(bootstrap)
        }

    def _generate_overrides(self, *args):
        # An override is a tuple of label, dictionary
        # we cannot use a dictionary because we need to preserve
        # the order of the arguments passed in as they are significant
        overrides = []

        for index, override in enumerate(args):
            source = self._extract_source(index, override)
            overrides.append(source.generate_override(self._separator))

        # We manually generate defaults here so that it is easy to find out
        # if there are defaults for fallbacks that should be applied.
        overrides.append(
            (
                '__defaults__',
                yapconf.flatten(self.defaults, separator=self._separator)
            )
        )
        return overrides

    def _explode_override(self, override):
        label = None
        unpacked_value = override
        file_type = self._file_type

        # If they provided a tuple, they may have named their tuple
        # or given us a differing file_type. So we unpack it before
        # we look at the unpacked value.
        if isinstance(override, tuple):
            if len(override) < 2:
                raise YapconfLoadError(
                    'Invalid override tuple provided. The tuple must have a '
                    'name and value. Optionally, it can provide a third '
                    'argument specifying the file type if it is different '
                    'than the default file_type passed during initialization.'
                )

            label = override[0]
            unpacked_value = override[1]
            if len(override) > 2:
                file_type = override[2]

        return label, unpacked_value, file_type

    def _extract_string_source(self, label, value, file_type):
        if value in self._sources:
            return self._sources[value]
        elif value == 'ENVIRONMENT':
            return get_source(value, 'environment')
        elif file_type in ['json', 'yaml']:
            return get_source(
                value,
                file_type,
                filename=value,
                encoding=self._encoding
            )

        raise YapconfLoadError(
            'Invalid override given: %s. A string type was detected '
            'but no valid source could be generated. This should be '
            'a string which points to a label from the sources you added, '
            'the string "ENVIRONMENT" or have a file_type of "json" or "yaml"'
            'got a (value, file_type) of: (%s %s)' %
            (label, value, file_type)
        )

    def _extract_source(self, index, override):
        label, unpacked_value, file_type = self._explode_override(override)

        if isinstance(unpacked_value, six.string_types):
            return self._extract_string_source(
                label, unpacked_value, file_type
            )

        elif isinstance(unpacked_value, dict):
            label = label or 'dict-%d' % index
            return get_source(label, 'dict', data=unpacked_value)

        raise YapconfLoadError(
            'Invalid override given: %s overrides must be one of the '
            'following: a dictionary, a filename, a label for a source, or '
            '"ENVIRONMENT". Got: %s' % (label, unpacked_value)
        )



# -*- coding: utf-8 -*-
import abc
import hashlib
import json
import os
import threading
import time

import six
from watchdog.observers import Observer

import yapconf
from yapconf.exceptions import YapconfLoadError, YapconfSourceError
from yapconf.handlers import FileHandler

if yapconf.kubernetes_support:
    from kubernetes.watch import watch

if yapconf.etcd_support:
    from etcd import EtcdWatchTimedOut


def get_source(label, source_type, **kwargs):
    """Get a config source based on type and keyword args.

    This is meant to be used internally by the spec via ``add_source``.

    Args:
        label (str): The label for this source.
        source_type: The type of source. See ``yapconf.SUPPORTED_SOURCES``

    Keyword Args:
        The keyword arguments are based on the source_type. Please see the
        documentation of the individual sources for a detailed list of all
        possible arguments.

    Returns (yapconf.sources.ConfigSource):
        A valid config source which can be used for generating an override.

    Raises:
        YapconfSourceError: If there is some kind of error with this source
        definition.

    """
    if source_type not in yapconf.ALL_SUPPORTED_SOURCES:
        raise YapconfSourceError(
            'Invalid source type %s. Supported types are %s.' %
            (source_type, yapconf.ALL_SUPPORTED_SOURCES)
        )
    if source_type not in yapconf.SUPPORTED_SOURCES:
        raise YapconfSourceError(
            'Unsupported source type "%s". If you want to use this type, you '
            'will need to install the correct client for it (try `pip install '
            'yapconf[%s]. Currently supported types are %s. All supported '
            'types are %s' %
            (source_type, source_type, yapconf.SUPPORTED_SOURCES,
             yapconf.ALL_SUPPORTED_SOURCES)
        )

    # We pop arguments from kwargs because the individual config sources
    # have better error messages if a keyword argument is missed.
    if source_type == 'dict':
        return DictConfigSource(label, data=kwargs.get('data'))

    elif source_type == 'json':
        return JsonConfigSource(label, **kwargs)

    elif source_type == 'yaml':
        filename = kwargs.get('filename')
        if 'filename' in kwargs:
            kwargs.pop('filename')
        return YamlConfigSource(label, filename, **kwargs)

    elif source_type == 'environment':
        return EnvironmentConfigSource(label)

    elif source_type == 'etcd':
        return EtcdConfigSource(
            label, kwargs.get('client'), kwargs.get('key', '/')
        )

    elif source_type == 'kubernetes':
        name = kwargs.get('name')
        if 'name' in kwargs:
            kwargs.pop('name')

        client = kwargs.get('client')
        if 'client' in kwargs:
            kwargs.pop('client')
        return KubernetesConfigSource(label, client, name, **kwargs)

    else:
        raise NotImplementedError(
            'No implementation for source type %s' % source_type
        )


@six.add_metaclass(abc.ABCMeta)
class ConfigSource(object):
    """Base class for a configuration source.

    Config sources will be used to generate overrides during configuration
    loading. In later iteration, it will also be used to migrate configs
    based on the configuration type.

    The act of loading configurations/migrating those configurations and
    especially watching those configuration is complicated enough to warrant
    its own data structure.

    Attributes:
          label: The label for this config source.
    """

    def __init__(self, label):
        self.label = label
        self.validate()
        self._continue = True

    @abc.abstractmethod
    def validate(self):
        """Validate that this ConfigSource can be used."""
        pass

    def generate_override(self, separator='.'):
        """Generate an override.

        Uses ``get_data`` which is expected to be implemented by each child
        class.

        Returns:
            A tuple of label, dict

        Raises:
            YapconfLoadError: If a known error occurs.

        """
        data = self.get_data()
        if not isinstance(data, dict):
            raise YapconfLoadError(
                'Invalid source (%s). The data was loaded successfully, but '
                'the result was not a dictionary. Got %s' %
                (self.label, type(data))
            )
        return self.label, yapconf.flatten(data, separator=separator)

    def _spawn_and_watch(self, handler):
        thread = self._spawn_watcher(handler)
        while self._continue:
            if thread.isAlive():
                time.sleep(5)
            else:
                thread = self._spawn_watcher(handler)

    def _spawn_watcher(self, handler):
        thread = threading.Thread(
            name=self.label + '-watcher',
            target=self._watch,
            args=(handler, self.get_data()),
        )
        thread.setDaemon(True)
        return thread

    def watch(self, handler, eternal=False):
        """Watch a source for changes. When changes occur, call the handler.

        By default, watches a dictionary that is in memory.

        Args:
            handler: Must respond to handle_config_change
            eternal: Spawn eternal watch, or just a single watch.

        Returns:
            The daemon thread that was spawned.

        """
        if eternal:
            thread = threading.Thread(
                name=self.label + '-watcher-eternal',
                target=self._spawn_and_watch,
                args=(handler,)
            )
            thread.setDaemon(True)
        else:
            thread = self._spawn_watcher(handler)
        thread.start()
        return thread

    def _hash_data(self, data):
        return hashlib.md5(repr(data).encode('utf-8')).hexdigest()

    def _watch(self, handler, data):
        current_hash = self._hash_data(data)
        while self._continue:
            new_hash = self._hash_data(self.get_data())
            if current_hash != new_hash:
                handler.handle_config_change(self.data)
                current_hash = new_hash
            time.sleep(1)

    @abc.abstractmethod
    def get_data(self):
        pass


class DictConfigSource(ConfigSource):
    """A basic config source with just a dictionary as the data.

    Keyword Args:
        data (dict): A dictionary that represents the data.
    """

    def __init__(self, label, data):
        self.data = data
        self.type = 'dict'
        super(DictConfigSource, self).__init__(label)

    def validate(self):
        if not isinstance(self.data, dict):
            raise YapconfSourceError(
                'Invalid source provided. %s provided data but it was not a '
                'dictionary. Got: %s' % (self.label, self.data)
            )

    def get_data(self):
        return self.data


class JsonConfigSource(ConfigSource):
    """JSON Config source.

    Needs either a filename or data keyword arg to work.

    Keyword Args:
        data (str): If provided, will be loaded via ``json.loads``
        filename (str): If provided, will be loaded via ``yapconf.load_file``
        kwargs: All other keyword arguments will be provided as keyword args
        to the ``load`` calls above.
    """

    def __init__(self, label, data=None, filename=None, **kwargs):
        self.type = 'json'
        self.data = data
        self.filename = filename
        self._load_kwargs = kwargs
        if 'encoding' not in self._load_kwargs:
            self._load_kwargs['encoding'] = 'utf-8'
        super(JsonConfigSource, self).__init__(label)

    def validate(self):
        if self.data is None and self.filename is None:
            raise YapconfSourceError(
                'Invalid source (%s). No data or filename was provided for a '
                'JSON config source.' % self.label
            )

    def _watch(self, handler, data):
        if self.filename:
            file_handler = FileHandler(
                filename=self.filename,
                handler=handler,
                file_type='json'
            )
            observer = Observer()
            directory = os.path.dirname(self.filename)
            observer.schedule(file_handler, directory, recursive=False)
            observer.start()
            observer.join()
        else:
            raise YapconfSourceError(
                'Cannot watch a string json source. Strings are immutable.'
            )

    def get_data(self):
        if self.data is not None:
            return json.loads(self.data, **self._load_kwargs)

        return yapconf.load_file(
            self.filename,
            file_type='json',
            klazz=YapconfSourceError,
            load_kwargs=self._load_kwargs
        )


class YamlConfigSource(ConfigSource):
    """YAML Config source.

    Needs a filename to work.

    Keyword Args:
        filename (str): Will be loaded via ``yapconf.load_file``
        encoding (str): The encoding of the filename.
    """

    def __init__(self, label, filename, **kwargs):
        self.type = 'yaml'
        self.filename = filename
        self._encoding = kwargs.get('encoding', 'utf-8')
        super(YamlConfigSource, self).__init__(label)

    def validate(self):
        if self.filename is None:
            raise YapconfSourceError(
                'Invalid source (%s). No filename was provided for a YAML '
                'config source.' % self.label
            )

    def _watch(self, handler, data):
        file_handler = FileHandler(
            filename=self.filename,
            handler=handler,
            file_type='yaml'
        )
        observer = Observer()
        directory = os.path.dirname(self.filename)
        observer.schedule(file_handler, directory, recursive=False)
        observer.start()
        observer.join()

    def get_data(self):
        return yapconf.load_file(
            self.filename,
            file_type='yaml',
            klazz=YapconfSourceError,
            open_kwargs={'encoding': self._encoding}
        )


class EnvironmentConfigSource(DictConfigSource):
    """Special dict config which gets its value from the environment."""

    def __init__(self, label):
        super(EnvironmentConfigSource, self).__init__(label, os.environ.copy())
        self.type = 'environment'

    def get_data(self):
        return os.environ.copy()


class EtcdConfigSource(ConfigSource):
    """Etcd config source (requires python-etcd package).

    If your keys have '/'s in them, you're going to have a bad time.

    Keyword Args:
        client: An etcd client from the python-etcd package.
        key (str): The key to fetch in etcd. Defaults to "/"
    """

    def __init__(self, label, client, key='/'):
        self.type = 'etcd'
        self.client = client
        self.key = key
        super(EtcdConfigSource, self).__init__(label)

    def validate(self):
        if not isinstance(self.client, yapconf.etcd_client.Client):
            raise YapconfSourceError(
                'Invalid source (%s). Client must be supplied and must be of '
                'type %s. Got type: %s' %
                (self.label,
                 type(yapconf.etcd_client.Client),
                 type(self.client))
            )

    def get_data(self):
        result = self.client.read(key=self.key, recursive=True)
        if not result.dir:
            raise YapconfLoadError(
                'Loaded key %s, but it was not a directory according to etcd.'
                % self.key
            )

        data = {}
        for child in result.children:
            keys = self._extract_keys(child.key)
            self._add_value(data, keys, child.value)

        return data

    def _watch(self, handler, data):
        while self._continue:
            try:
                self.client.read(key=self.key, wait=True, recursive=True)
                handler.handle_config_change(self.get_data())
            except EtcdWatchTimedOut:
                pass

    def _add_value(self, data, keys, value):
        for i, key in enumerate(keys):
            if i == len(keys) - 1:
                data[key] = value
            else:
                data[key] = {}

    def _extract_keys(self, fq_key):
        return fq_key.lstrip(self.key).split('/')


class KubernetesConfigSource(ConfigSource):
    """A kubernetes config data source.

    This is meant to load things directly from the kubernetes API.
    Specifically, it can load things from config maps.

    Keyword Args:
        client: A kubernetes client from the kubernetes package.
        name (str): The name of the ConfigMap to load.
        namespace (str): The namespace for the ConfigMap
        key (str): A key for the given ConfigMap data object.
        config_type (str): Used in conjunction with 'key', if 'key' points to
        a data blob, this will specify whether to use json or yaml to load
        the file.
    """

    def __init__(self, label, client, name, **kwargs):
        self.type = 'kubernetes'
        self.client = client
        self.name = name
        self.namespace = kwargs.get('namespace') or 'default'
        self.key = kwargs.get('key')
        self.config_type = kwargs.get('config_type') or 'json'
        super(KubernetesConfigSource, self).__init__(label)

    def validate(self):
        if (yapconf.kubernetes_client and
                not isinstance(self.client,
                               yapconf.kubernetes_client.CoreV1Api)):
            raise YapconfSourceError(
                'Invalid source (%s). Client must be supplied and must be of '
                'type %s. Got type: %s' %
                (self.label,
                 type(yapconf.kubernetes_client.CoreV1Api),
                 type(self.client))
            )

        if self.config_type == 'yaml' and not yapconf.yaml_support:
            raise YapconfSourceError(
                'Kubernetes config source specified that the configmap '
                'contained a yaml config. In order to support yaml loading, '
                'you will need to `pip install yapconf[yaml]`.'
            )

        if self.config_type not in yapconf.FILE_TYPES:
            raise YapconfSourceError(
                'Invalid config type specified for a kubernetes config. %s is '
                'not supported. Supported types are %s' %
                (self.name, yapconf.FILE_TYPES)
            )

    def get_data(self):
        result = self.client.read_namespaced_config_map(self.name,
                                                        self.namespace)
        if self.key is None:
            return result.data

        if self.key not in result.data:
            raise YapconfLoadError(
                'Loaded configmap %s, but could not find key %s in the data '
                'portion of the configmap.' % self.key
            )

        nested_config = result.data[self.key]
        if self.config_type == 'json':
            return json.loads(nested_config, )
        elif self.config_type == 'yaml':
            return yapconf.yaml.load(nested_config)
        else:
            raise NotImplementedError('Cannot load config with type %s' %
                                      self.config_type)

    def _watch(self, handler, data):
        w = watch.Watch()
        for event in w.stream(
            self.client.list_namespaced_config_map,
            namespace=self.namespace
        ):
            config_map = event['object']
            if config_map.metadata.name != self.name:
                continue

            if event['type'] == 'DELETED':
                raise YapconfSourceError(
                    'Kubernetes config watcher died. The configmap %s was '
                    'deleted.' % self.name
                )

            if event['type'] == 'MODIFIED':
                handler.handle_config_change(self.get_data())



# -*- coding: utf-8 -*-
import sys
from argparse import ArgumentParser

import pytest

import yapconf
from yapconf.exceptions import (YapconfDictItemError, YapconfItemError,
                                YapconfItemNotFound, YapconfListItemError,
                                YapconfValueError)
from yapconf.items import (YapconfBoolItem, YapconfDictItem, YapconfItem,
                           YapconfListItem, from_specification)

if sys.version_info > (3,):
    long = int


@pytest.mark.parametrize('clazz,kwargs,env_names', [
    (
        YapconfItem,
        {'name': 'foo', 'env_name': None, 'alt_env_names': []},
        ['FOO']
    ),
    (
        YapconfItem,
        {'name': 'foo', 'env_name': None, 'alt_env_names': ['bar']},
        ['FOO', 'bar']
    ),
    (
        YapconfItem,
        {'name': 'foo', 'env_name': None, 'alt_env_names': ['bar'],
         'env_prefix': 'MY_APP_'},
        ['MY_APP_FOO', 'MY_APP_bar']
    ),
    (
        YapconfListItem,
        {
            'name': 'foos',
            'env_name': 'FOOS',
            'alt_env_names': ['BARS'],
            'children': {'foo': YapconfItem('foo')}
        },
        []
    )
])
def test_env_names(clazz, kwargs, env_names):
    item = clazz(**kwargs)
    item = clazz(**kwargs)
    assert item.all_env_names == env_names


@pytest.mark.parametrize('clazz,kwargs,error_clazz', [
    (
        YapconfItem,
        {'name': 'foo', 'item_type': 'INVALID_TYPE'},
        YapconfItemError
    ),
    (
        YapconfItem,
        {'name': 'foo', 'choices': ['a'], 'default': 'b'},
        YapconfValueError
    ),
    (
        YapconfItem,
        {'name': 'foo', 'cli_short_name': 'too_long'},
        YapconfItemError
    ),
    (
        YapconfItem,
        {'name': 'foo', 'cli_short_name': '-'},
        YapconfItemError
    ),
    (
        YapconfDictItem,
        {'name': 'foo', 'children': {}},
        YapconfDictItemError
    ),
    (
        YapconfDictItem,
        {
            'name': 'foo',
            'children': {'item1': YapconfItem('simple_item')},
            'choices': [{'foo': 'bar'}]
        },
        YapconfDictItemError
    ),
    (
        YapconfListItem,
        {
            'name': 'foo',
            'children': {
                'item1': YapconfItem('simple_item'),
                'item2': YapconfBoolItem('my_bool')}
        },
        YapconfListItemError
    )
])
def test_bad_specifications(clazz, kwargs, error_clazz):
    with pytest.raises(error_clazz):
        clazz(**kwargs)


def test_get_config_value_not_found_and_required(simple_item):
    with pytest.raises(YapconfItemNotFound):
        simple_item.get_config_value({})


def test_get_config_value_not_required(simple_item):
    simple_item.required = False
    assert simple_item.get_config_value({}) is None


@pytest.mark.parametrize('default,config,expected', [
    ('default', [('dict1', {})], 'default'),
    ('default', [('dict1', {'foo': None})], 'default')
])
def test_get_config_value_from_default(simple_item, default, config, expected):
    simple_item.default = default
    assert simple_item.get_config_value(config) == expected


def test_get_config_value_from_override(simple_item):
    simple_item.default = 'default'
    assert simple_item.get_config_value([('dict-1', {'foo': 'bar'})]) == 'bar'


@pytest.mark.parametrize('item_type,orig,expected', [
    ('str', 123, '123'),
    ('str', u'🐍', u'🐍'),
    ('int', '123', 123),
    ('long', '123', long('123')),
    ('complex', '2j', 2j),
])
def test_convert_config_value(simple_item, item_type, orig, expected):
    simple_item.item_type = item_type
    assert simple_item.convert_config_value(orig, 'label') == expected


@pytest.mark.parametrize('orig,expected', [
    (True, True),
    (False, False),
    ('t', True),
    ('TrUe', True),
    ('y', True),
    ('yes', True),
    ('f', False),
    ('False', False),
    ('n', False),
    ('no', False),
    (1, True),
    (0, False),
    ('1', True),
    ('0', False),
])
def test_convert_bool_config_value(bool_item, orig, expected):
    assert bool_item.convert_config_value(orig, 'label') == expected


@pytest.mark.parametrize('item_type,value,exc_clazz', [
    ('int', [], YapconfValueError),
    ('INVALID_TYPE', 'value', YapconfItemError),
    ('int', object, YapconfValueError),
])
def test_convert_config_value_error(simple_item, item_type, value, exc_clazz):
    simple_item.item_type = item_type
    with pytest.raises(exc_clazz):
        simple_item.convert_config_value(value, 'label')


def test_convert_config_value_invalid_bool(bool_item):
    with pytest.raises(YapconfValueError):
        bool_item.convert_config_value('INVALID_VALUE', 'label')


@pytest.mark.parametrize('env_name,alt_names, default,config,expected', [
    ('FOO', [], None, {'FOO': 'foo_value', 'foo': 'should not be this'},
     'foo_value'),
    ('FOO', [], 'default', {'FOO': 'foo_value', 'foo': 'should not be this'},
     'foo_value'),
    ('FOO', [], 'default', {'FOO': None}, 'default'),
    ('FOO', [], 'default', {'FOO': ''}, 'default'),
    ('FOO', ['FOO_BACKUP'], None, {'FOO_BACKUP': 'foo_backup_value'},
     'foo_backup_value')
])
def test_get_config_value_from_environment(simple_item, env_name, alt_names,
                                           default, config, expected):
    simple_item.alt_env_names = alt_names
    simple_item.env_name = env_name
    simple_item.default = default
    value = simple_item.get_config_value([('ENVIRONMENT', config)])
    assert value == expected


def test_get_config_value_unformatted_env(unformatted_item):
    value = unformatted_item.get_config_value(
        [('ENVIRONMENT', {unformatted_item.name: 'value'})])
    assert value == 'value'


def test_get_config_value_for_list(list_item):
    value = list_item.get_config_value([('label', {'foos': ['foo1', 'foo2']})])
    assert value == ['foo1', 'foo2']


def test_get_config_value_for_list_not_there(list_item):
    with pytest.raises(YapconfItemNotFound):
        list_item.get_config_value({})


def test_get_config_value_with_default(list_item):
    list_item.default = ['foo1', 'foo2']
    value = list_item.get_config_value({})
    assert value == ['foo1', 'foo2']


def test_get_config_list_value_not_required(list_item):
    list_item.required = False
    assert list_item.get_config_value({}) is None


def test_get_config_value_not_a_list(list_item):
    with pytest.raises(YapconfValueError):
        list_item.get_config_value([('label', {'foos': 123})])


def test_convert_config_value_not_a_list(list_item):
    with pytest.raises(YapconfValueError):
        list_item.convert_config_value(123, 'label')


def test_get_config_ignore_environment(list_item):
    list_item.env_name = 'FOOS'
    value = list_item.get_config_value(
        [
            ('ENVIRONMENT', {'FOOS': ['foo1', 'foo2']}),
            ('label', {'foos': ['foo3', 'foo4']})
        ]
    )
    assert value == ['foo3', 'foo4']


def test_dict_get_config_value(dict_item):
    value = dict_item.get_config_value(
        [
            ('label1', yapconf.flatten({'foo_dict': {'foo': 'bar'}}))
        ]
    )
    assert value == {'foo': 'bar'}


def test_dict_get_config_value_ignore_environment(dict_item):
    dict_item.env_name = 'FOO_DICT'
    value = dict_item.get_config_value([
        ('ENVIRONMENT', {'FOO_DICT': {'foo': 'bar'}}),
        ('label1', yapconf.flatten({'foo_dict': {'foo': 'baz'}}))
    ])
    assert value == {'foo': 'baz'}


@pytest.mark.parametrize('current_default,new_default,respect_flag,expected', [
    (None, 'foo', False, 'foo'),
    (None, None, False, None),
    ('foo', None, False, 'foo'),
    ('foo', None, True, None)
])
def test_update_default(simple_item, current_default, new_default,
                        respect_flag, expected):
    simple_item.default = current_default
    simple_item.update_default(new_default, respect_none=respect_flag)
    assert simple_item.default == expected


@pytest.mark.parametrize('type,required,default,short,name,args,expected', [
    (
        'str',
        True,
        'default_value',
        None,
        None,
        ['--foo', 'foo_value'],
        'foo_value'
    ),
    ('str', True, 'default_value', None, None, [], None),
    ('str', False, None, None, None, [], None),
    ('str', True, None, 'f', None, ['-f', 'foo_value'], 'foo_value'),
    ('str', True, None, 'f', 'alt', ['--alt', 'foo_value'], 'foo_value'),
    ('int', True, None, 'f', None, ['-f', '1'], 1),
    ('long', True, None, 'f', None, ['-f', '1'], long(1)),
    ('float', True, None, 'f', None, ['-f', '1.23'], 1.23),
    ('complex', True, None, 'f', None, ['-f', '2j'], 2j),
])
def test_add_argument(
    simple_item,
    type,
    required,
    default,
    short,
    name,
    args,
    expected
):
    simple_item.item_type = type
    simple_item.required = required
    simple_item.default = default
    simple_item.cli_short_name = short
    simple_item.cli_name = name
    parser = ArgumentParser()
    simple_item.add_argument(parser)
    values = vars(parser.parse_args(args))
    assert values[simple_item.name] == expected


@pytest.mark.parametrize('default,short_name,cli_name,args,expected', [
    (True, None, None, ['--no-my-bool'], False),
    (False, None, None, ['--my-bool'], True),
    (None, None, None, ['--my-bool'], True),
    (None, None, None, ['--no-my-bool'], False),
    (None, None, 'alt', ['--no-alt'], False),
    (None, None, 'alt', ['--alt'], True),
    (None, 'b', None, ['--no-b'], False),
    (None, 'b', None, ['--b'], True),
    (True, None, None, [], None),
    (False, None, None, [], None),
])
def test_add_bool_argument(
    bool_item, default, short_name, cli_name, args, expected
):
    bool_item.default = default
    bool_item.cli_short_name = short_name
    bool_item.cli_name = cli_name
    parser = ArgumentParser()
    bool_item.add_argument(parser)
    values = vars(parser.parse_args(args))
    assert values[bool_item.name] == expected


def test_add_argument_unformatted(unformatted_item):
    parser = ArgumentParser()
    unformatted_item.add_argument(parser)
    args = ['--%s' % unformatted_item.name, 'value']
    values = vars(parser.parse_args(args))
    assert values[unformatted_item.name] == 'value'


def test_add_argument_unformatted_bool(unformatted_bool_item):
    parser = ArgumentParser()
    unformatted_bool_item.add_argument(parser)
    args = ['--no-%s' % unformatted_bool_item.name]
    values = vars(parser.parse_args(args))
    assert values[unformatted_bool_item.name] is False


@pytest.mark.parametrize('list_default,child_default,args,expected', [
    (None, True, ['--my-bool', '--no-my-bool'], [True, False]),
    (None, False, ['--no-my-bool', '--my-bool'], [False, True]),
    ([True, False], False, [], None),
    ([True, False], None, [], None),
    ([True, False], None, ['--my-bool'], [True]),
    ([True, False], None, ['--no-my-bool'], [False]),
])
def test_add_list_boolean_arguments(bool_list_item,
                                    list_default,
                                    child_default,
                                    args,
                                    expected):
    bool_list_item.default = list_default
    bool_list_item.child.default = child_default
    parser = ArgumentParser()
    bool_list_item.add_argument(parser)
    values = vars(parser.parse_args(args))

    assert values[bool_list_item.name] == expected


@pytest.mark.parametrize('list_default,child_default,args,expected', [
    (None, None, ['--foo', 'foo1', '--foo', 'foo2'], ['foo1', 'foo2']),
    (None, "foo", ['--foo', 'foo1', '--foo', 'foo2'], ['foo1', 'foo2']),
    (['foo1', 'foo2'], 'foo', [], None),
    ([], 'foo', [], None),
    (['foo1', 'foo2'], None, ['--foo', 'foo'], ['foo']),
])
def test_add_list_arguments(list_item,
                            list_default,
                            child_default,
                            args,
                            expected):
    list_item.default = list_default
    list_item.child.default = child_default
    parser = ArgumentParser()
    list_item.add_argument(parser)
    values = vars(parser.parse_args(args))
    assert values[list_item.name] == expected


@pytest.mark.parametrize('dict_default,args,expected', [
    (
        None,
        ['--db-name', 'db_name', '--db-port', '123', '--db-verbose',
         '--db-users', 'user1', '--db-users', 'user2',
         '--db-log-level', 'INFO',
         '--db-log-file', '/path/to/file'],
        {
            'name': 'db_name',
            'port': 123,
            'verbose': True,
            'users': ['user1', 'user2'],
            'log': {
                'level': 'INFO',
                'file': '/path/to/file'
            }
        }
    )
])
def test_basic_dict_add_argument(db_item,
                                 dict_default,
                                 args,
                                 expected):
    db_item.default = dict_default
    parser = ArgumentParser()
    db_item.add_argument(parser)
    values = vars(parser.parse_args(args))
    assert values[db_item.name] == expected


def test_basic_dict_get_config_value_from_env(db_item):
    value = db_item.get_config_value(
        [
            (
                'ENVIRONMENT',
                {
                    'ALT_NAME': 'db_name',
                    'DB_PORT': '1234',
                    'DB_VERBOSE': 'True'
                }
            ),
            (
                'dict1',
                yapconf.flatten({
                    'db': {
                        'users': ['user1', 'user2'],
                        'log': {'level': 'INFO', 'file': '/some/file'}
                    }
                })
            )
        ]
    )
    assert value['name'] == 'db_name'
    assert value['port'] == 1234


@pytest.mark.parametrize('item,config', [
    (
        YapconfItem("foo", choices=['a', 'b', 'c']),
        {'foo': 'd'}
    ),
    (
        YapconfListItem("foos",
                        children={
                            'foo': YapconfItem('foo',
                                               choices=['a', 'b', 'c'])
                        }),
        {'foos': ['d']},
    ),
    (
        YapconfListItem("foos",
                        choices=[['a', 'a'], ['a', 'b'], ['c']],
                        children={
                            'foo': YapconfItem('foo',
                                               choices=['a', 'b', 'c'])
                        }),
        {'foos': ['a']}
    ),
    (
        YapconfDictItem("foo_dict",
                        children={
                            "foo": YapconfItem('foo',
                                               choices=['bar', 'baz'],
                                               prefix='foo_dict')
                        }),
        yapconf.flatten({'foo_dict': {'foo': 'invalid'}})
    ),
])
def test_choices(item, config):
    with pytest.raises(YapconfValueError):
        item.get_config_value([('test', config)])


def test_from_spec_list_fq_names():
    spec = {
        'list1': {
            'type': 'list',
            'required': True,
            'items': {
                'list_item': {
                    'type': 'str',
                    'required': True
                },
            },
        },
        'list2': {
            'type': 'list',
            'required': True,
            'items': {
                'list_dict_item': {
                    'type': 'dict',
                    'required': True,
                    'items': {
                        'foo': {},
                        'bar': {},
                    },
                },
            },
        },
    }
    items = from_specification(spec)
    assert items['list1'].fq_name == 'list1'
    assert items['list1'].children['list1'].fq_name == 'list1'
    assert items['list2'].children['list2'].fq_name == 'list2'
    assert (
        items['list2'].children['list2'].children['foo'].fq_name ==
        'list2.foo'
    )
    assert (
        items['list2'].children['list2'].children['bar'].fq_name ==
        'list2.bar'
    )


def test_from_specification_check_fq_names():
    spec = {
        'foo': {
            'type': 'dict',
            'required': True,
            'items': {
                'bar': {
                    'type': 'dict',
                    'required': True,
                    'items': {
                        'baz': {
                            'type': 'str', 'required': True,
                        },
                    },
                },
                'bat': {
                    'type': 'bool',
                }
            },
        },
    }
    items = from_specification(spec)
    assert items['foo'].fq_name == 'foo'
    assert items['foo'].children['bar'].fq_name == 'foo.bar'
    assert (
        items['foo'].children['bar'].children['baz'].fq_name == 'foo.bar.baz'
    )
    assert items['foo'].children['bat'].fq_name == 'foo.bat'


def test_invalid_value():
    def always_invalid(value):
        return False
    item = from_specification({'foo': {'validator': always_invalid}})['foo']
    with pytest.raises(YapconfValueError):
        item.get_config_value([('label', {'foo': 'bar'})])


def test_filter():
    item = YapconfItem("name")
    prefix_item = YapconfItem("name", prefix="prefix")

    assert item.apply_filter(include=["name"]) is item
    assert item.apply_filter(exclude=["name"]) is None
    assert item.apply_filter(bootstrap=False) is item
    assert item.apply_filter(bootstrap=True) is None
    assert item.apply_filter(exclude_bootstrap=True) is item
    assert item.apply_filter(exclude_bootstrap=False) is item
    assert item.apply_filter() is item

    assert prefix_item.apply_filter(include=["name"]) is None
    assert prefix_item.apply_filter(include=["prefix.name"]) is prefix_item


def test_dict_filter(db_item):
    assert db_item.apply_filter(include=["db"]) is db_item

    actual = db_item.apply_filter(include=["db.name"])
    assert actual is not None
    assert len(actual.children) == 1
    assert actual.children["name"].fq_name == "db.name"

    actual = db_item.apply_filter(include=["db.name", "db.log.level"])
    assert len(actual.children) == 2
    assert "name" in actual.children
    assert "log" in actual.children
    assert len(actual.children["log"].children) == 1
    assert "level" in actual.children["log"].children

    actual = db_item.apply_filter(
        include=["db.name", "db.log.level"], exclude=["db.log.file"]
    )
    assert len(actual.children) == 2
    assert "name" in actual.children
    assert "log" in actual.children
    assert len(actual.children["log"].children) == 1
    assert "level" in actual.children["log"].children

    actual = db_item.apply_filter(
        exclude=["db.log"]
    )
    assert len(actual.children) == 4
    assert "log" not in actual.children

    actual = db_item.apply_filter(
        exclude=["db.log.file"]
    )
    assert len(actual.children) == 5
    assert "log" in actual.children
    assert "file" not in actual.children["log"].children



# -*- coding: utf-8 -*-
import pytest
from etcd import EtcdWatchTimedOut
from mock import Mock, PropertyMock, patch

import yapconf
from yapconf.exceptions import YapconfSourceError
from yapconf.sources import (DictConfigSource, EnvironmentConfigSource,
                             EtcdConfigSource, JsonConfigSource,
                             KubernetesConfigSource, YamlConfigSource,
                             get_source)

original_sources = yapconf.SUPPORTED_SOURCES.copy()


def teardown_function(function):
    yapconf.SUPPORTED_SOURCES = original_sources


@pytest.mark.parametrize('source_type,sources,kwargs', [
    ('invalid_source', yapconf.SUPPORTED_SOURCES, {}),
    ('yaml', [], {}),
    ('dict', yapconf.SUPPORTED_SOURCES, {'data': 'NOT_A_DICT'}),
    ('json', yapconf.SUPPORTED_SOURCES, {}),
    ('yaml', yapconf.SUPPORTED_SOURCES, {}),
    ('etcd', yapconf.SUPPORTED_SOURCES, {'client': 'NOT_A_ETCD_CLIENT'}),
    ('kubernetes', yapconf.SUPPORTED_SOURCES, {'client': 'NOT_A_K8S_CLIENT'}),
])
def test_get_source_error(source_type, sources, kwargs):
    yapconf.SUPPORTED_SOURCES = sources
    with pytest.raises(YapconfSourceError):
        get_source('label', source_type, **kwargs)


@pytest.mark.parametrize('source_type,kwargs,klazz', [
    ('dict', {'data': {}}, DictConfigSource),
    ('json', {'filename': 'path/to/config'}, JsonConfigSource),
    ('yaml', {'filename': 'path/to/config'}, YamlConfigSource),
    ('environment', {}, EnvironmentConfigSource),
    (
        'etcd',
        {'client': Mock(spec=yapconf.etcd_client.Client)},
        EtcdConfigSource),
    (
        'kubernetes',
        {
            'name': 'configMap',
            'client': Mock(spec=yapconf.kubernetes_client.CoreV1Api)
        },
        KubernetesConfigSource
    ),
])
def test_get_source(source_type, kwargs, klazz):
    source = get_source('label', source_type, **kwargs)
    assert isinstance(source, klazz)


def test_k8s_watch():
    def apply_name(mock, name):
        p = PropertyMock(return_value=name)
        type(mock).name = p

    md_mock1 = Mock()
    apply_name(md_mock1, 'name')
    event_object1 = Mock(metadata=md_mock1)

    events = [
        {
            'object': event_object1,
            'type': 'MODIFIED'
        },
    ]

    mock_client = Mock(spec=yapconf.kubernetes_client.CoreV1Api)
    handler = Mock()
    source = get_source('label', 'kubernetes', client=mock_client, name='name')
    source.get_data = Mock(return_value="NEW DATA")
    with patch('yapconf.sources.watch.Watch') as WatchMock:
        mock_watch = Mock()
        WatchMock.return_value = mock_watch

        mock_watch.stream = Mock(return_value=events)
        source._watch(handler, {})

    handler.handle_config_change.assert_called_with("NEW DATA")


def test_k8s_watch_bomb_on_deletion():
    def apply_name(mock, name):
        p = PropertyMock(return_value=name)
        type(mock).name = p

    md_mock1 = Mock()
    apply_name(md_mock1, 'DOES_NOT_MATCH')
    event_object1 = Mock(metadata=md_mock1)

    md_mock2 = Mock()
    apply_name(md_mock2, 'name')
    event_object2 = Mock(metadata=md_mock2)
    events = [
        {
            'object': event_object1,
            'type': 'DELETED'
        },
        {
            'object': event_object2,
            'type': 'DELETED'
        }
    ]

    mock_client = Mock(spec=yapconf.kubernetes_client.CoreV1Api)
    mock_client.list_namespaced_config_map = Mock(return_value=[])
    handler = Mock()
    source = get_source('label', 'kubernetes', client=mock_client, name='name')
    with patch('yapconf.sources.watch.Watch') as WatchMock:
        mock_watch = Mock()
        WatchMock.return_value = mock_watch

        mock_watch.stream = Mock(return_value=events)
        with pytest.raises(YapconfSourceError):
            source._watch(handler, {})


def test_etcd_watch():

    mock_client = Mock(
        spec=yapconf.etcd_client.Client,
        read=Mock(side_effect=[EtcdWatchTimedOut, "result", ValueError])
    )
    handler = Mock()
    source = get_source('label', 'etcd', client=mock_client)
    source.get_data = Mock(return_value='NEW_DATA')
    with pytest.raises(ValueError):
        source._watch(handler, {})

    handler.handle_config_change('NEW_DATA')


def test_environment():
    source = get_source('label', 'environment')
    assert isinstance(source.get_data(), dict)
